@@ -4,7 +4,7 @@
 
 ;; Author: Jim Porter
 ;; URL: https://github.com/jimporter/urgrep
-;; Version: 0.2.0
+;; Version: 0.3.0
 ;; Keywords: grep, search
 ;; Package-Requires: ((emacs "27.1") (compat "29.1.0.1") (project "0.3.0"))
 
@@ -120,7 +120,7 @@ The currently-used tool can be inspected from the hook via
 
 ;; Urgrep utility functions
 
-;; `format-prompt' was added in Emacs 28.1.
+;; `split-string-shell-command' was added in Emacs 28.1.
 (defalias 'urgrep--split-string-shell-command
   (if (fboundp 'split-string-shell-command)
       #'split-string-shell-command
@@ -187,6 +187,25 @@ for MS shells."
       argument
     (shell-quote-argument argument)))
 
+(defun urgrep--shell-join (arguments)
+  "Join ARGUMENTS by spaces, quoting each argument as needed for the shell."
+  (mapconcat #'urgrep--maybe-shell-quote-argument arguments " "))
+
+(defun urgrep--safe-file-name (file)
+  "Get a safe form of FILE ready to pass to an external process.
+This expands tildes and removes any remote host identifiers or quoting."
+  (when-let ((remote (file-remote-p file)))
+    (unless (equal remote (file-remote-p default-directory))
+      (error "remote file doesn't match host for `default-directory'"))
+    (setq file (file-local-name file)))
+  (setq file (file-name-unquote file))
+  (if (not (string-prefix-p "~" file))
+      file
+    (setq file (expand-file-name file))
+    (if (file-in-directory-p file default-directory)
+        (file-relative-name file)
+      (file-local-name file))))
+
 (defun urgrep--flatten-arguments (tree &optional abbrs)
   "Flatten a TREE of arguments into a single shell-quoted string.
 This also finds sublists with the `:abbreviate' key and adds the
@@ -201,8 +220,7 @@ one for each `:abbreviate' key found."
           (while (consp elem)
             (when (eq (car elem) :abbreviate)
               (push (propertize
-                     (mapconcat #'urgrep--maybe-shell-quote-argument
-                                (flatten-list (cdr elem)) " ")
+                     (urgrep--shell-join (flatten-list (cdr elem)))
                      'abbreviated-command (or (pop abbrs) t))
                     elems)
               (throw 'abbreviated t))
@@ -212,6 +230,20 @@ one for each `:abbreviate' key found."
     (when tree (push (urgrep--maybe-shell-quote-argument tree) elems))
     (string-join (nreverse elems) " ")))
 
+(defun urgrep--interpolate-arguments (query tool props)
+  "Interpolate search arguments for QUERY using TOOL.
+PROPS is an alist of extra properties corresponding to the
+properties defined in the `urgrep-tools' entry for TOOL."
+  (let* ((arguments (urgrep--get-prop 'arguments tool))
+         (abbrev (urgrep--get-prop 'abbreviations tool))
+         (props `((executable . ,(urgrep--get-prop 'executable-name tool))
+                  (query . ,query)
+                  ,@(mapcar (pcase-lambda (`(,k . ,v))
+                              (cons k (urgrep--get-prop-pcase
+                                       k tool v "-arguments")))
+                            props))))
+    (urgrep--flatten-arguments (cl-sublis props arguments) abbrev)))
+
 (rx-define urgrep-regular-number (seq (any "1-9") (* digit)))
 
 
@@ -224,87 +256,93 @@ one for each `:abbreviate' key found."
     ((or `(,c . ,c) (and c (pred numberp))) (list (format "-C%d" c)))
     (`(,b . ,a) (list (format "-B%d" b) (format "-A%d" a)))))
 
-(cl-defun urgrep--rgrep-command (query &key tool regexp case-fold hidden files
-                                       directory context color
+(defun urgrep--add-grep-options (template extra-opts)
+  "Add EXTRA-OPTS to the specified GREP template and return it.
+This function changes match data."
+  (unless (string-match "<C>" template)
+    (error "grep template should have a <C> placeholder"))
+  ;; Locally add options to the template that grep.el isn't aware of.
+  (replace-match (concat "<C> " (urgrep--shell-join extra-opts))
+                 t t template))
+
+(cl-defun urgrep--rgrep-command (query &key tool regexp case-fold hidden
+                                       file-wildcard root context color
                                        &allow-other-keys)
   "Get the command to run for QUERY when using rgrep.
-Optional keys TOOL, REGEXP, CASE-FOLD, HIDDEN, FILES, DIRECTORY,
-CONTEXT, and COLOR are as in `urgrep-command'."
+Optional keys TOOL, REGEXP, CASE-FOLD, HIDDEN, FILE-WILDCARD,
+ROOT, CONTEXT, and COLOR are as in `urgrep-command'."
   (grep-compute-defaults)
-  ;; Locally add options to `grep-find-template' that grep.el isn't aware of.
-  (let ((grep-find-template grep-find-template)
+  (let ((case-fold-search nil)
         (grep-highlight-matches (if color 'always nil))
-        (files (if files (mapconcat #'identity files " ") "*"))
-        (directory (when directory
-                     (mapconcat #'urgrep--maybe-shell-quote-argument
-                                directory " "))))
-    (pcase-dolist (`(,k . ,v) `((regexp    . ,regexp)
-                                (case-fold . ,case-fold)
-                                (context   . ,context)))
-      (when-let ((args (urgrep--get-prop-pcase k tool v "-arguments"))
-                 (args (mapconcat #'urgrep--maybe-shell-quote-argument args
-                                  " "))
-                 ((string-match "<C>" grep-find-template)))
-        (setq grep-find-template
-              (replace-match (concat "<C> " args) t t grep-find-template))))
-    (let ((case-fold-search nil)
-          (grep-find-ignored-directories grep-find-ignored-directories)
-          (grep-find-ignored-files grep-find-ignored-files))
-      (unless hidden
-        (setq grep-find-ignored-directories
-              (cons ".*" (seq-filter (lambda (s) (not (string-prefix-p "." s)))
-                                     grep-find-ignored-directories))
-              grep-find-ignored-files
-              (cons ".*" (seq-filter (lambda (s) (not (string-prefix-p "." s)))
-                                     grep-find-ignored-files))))
-      (let ((command (rgrep-default-command query files directory)))
-        (save-match-data
-          ;; Hide excessive part of rgrep command.
-          (when (string-match
-                 (rx bol "find " (group (*? nonl)) " " (or "-exec" "-print"))
-                 command)
-            (put-text-property (match-beginning 1) (match-end 1)
-                               'abbreviated-command t command)))
-        command))))
+        (extra-opts
+         (apply #'append
+                (mapcar (pcase-lambda (`(,k . ,v))
+                          (urgrep--get-prop-pcase k tool v "-arguments"))
+                        `((context   . ,context)
+                          (case-fold . ,case-fold)
+                          (regexp    . ,regexp))))))
+    (save-match-data
+      (if root
+          (let ((file-wildcard
+                 (if file-wildcard (string-join file-wildcard " ") "*"))
+                (dirs (urgrep--shell-join root))
+                (grep-find-template (urgrep--add-grep-options
+                                     grep-find-template extra-opts))
+                (grep-find-ignored-directories grep-find-ignored-directories)
+                (grep-find-ignored-files grep-find-ignored-files))
+            (unless hidden
+              ;; Ignore all dotfiles.
+              (let ((pred (lambda (s) (not (string-prefix-p "." s)))))
+                (setq grep-find-ignored-directories
+                      (cons ".*" (seq-filter
+                                  pred grep-find-ignored-directories))
+                      grep-find-ignored-files
+                      (cons ".*" (seq-filter pred grep-find-ignored-files)))))
+            (let ((command (rgrep-default-command query file-wildcard dirs)))
+              ;; Hide excessive part of rgrep command.
+              (when (string-match (rx bol "find " (group (*? nonl)) " "
+                                      (or "-exec" "-print"))
+                     command)
+                (put-text-property (match-beginning 1) (match-end 1)
+                                   'abbreviated-command t command))
+              command))
+        ;; No ROOT, so do a regular `grep'.
+        (grep-expand-template
+         (urgrep--add-grep-options grep-template extra-opts) query)))))
 
 (cl-defun urgrep--git-grep-command (query &key tool regexp case-fold hidden
-                                          files directory group context color)
+                                          file-wildcard root group context
+                                          color)
   "Get the command to run for QUERY when using git grep.
-Optional keys TOOL, REGEXP, CASE-FOLD, HIDDEN, FILES, DIRECTORY,
-CONTEXT, and COLOR are as in `urgrep-command'."
-  ;; XXX: This is very similar to the implementation in `urgrep-command', except
-  ;; that we do some extra work to generate pathspecs. Can we factor out some of
-  ;; this?
-  (let* ((arguments (urgrep--get-prop 'arguments tool))
-         (abbrev (urgrep--get-prop 'abbreviations tool))
-         (pathspecs
-          (if (and files directory)
-              (mapcan
-               (lambda (file)
-                 (mapcar (lambda (dir)
-                           (concat ":(glob)" (file-name-concat dir "**" file)))
-                         directory))
-               files)
-            (or files directory)))
-         (props `((executable . ,(urgrep--get-prop 'executable-name tool))
-                  (query . ,query)
-                  ,@(mapcar (pcase-lambda (`(,k . ,v))
-                              (cons k (urgrep--get-prop-pcase
-                                       k tool v "-arguments")))
-                            `((regexp         . ,regexp)
-                              (case-fold      . ,case-fold)
-                              (hidden-file    . ,hidden)
-                              (pathspec       . ,pathspecs)
-                              (group          . ,group)
-                              (context        . ,context)
-                              (color          . ,color))))))
-    (urgrep--flatten-arguments (cl-sublis props arguments)
-                               abbrev)))
+Optional keys TOOL, REGEXP, CASE-FOLD, HIDDEN, FILE-WILDCARD,
+ROOT, CONTEXT, and COLOR are as in `urgrep-command'."
+  (let ((pathspecs
+         (cond
+          ((equal root '("."))
+           file-wildcard)
+          ((and file-wildcard root)
+           (mapcan
+            (lambda (file)
+              (mapcar (lambda (dir)
+                        (concat ":(glob)" (file-name-concat dir "**" file)))
+                      root))
+            file-wildcard))
+          (file-wildcard
+           (error ":file-wildcard expects a non-nil :root"))
+          (t root))))
+    (urgrep--interpolate-arguments query tool
+                                   `((regexp      . ,regexp)
+                                     (case-fold   . ,case-fold)
+                                     (hidden-file . ,hidden)
+                                     (pathspec    . ,pathspecs)
+                                     (group       . ,group)
+                                     (context     . ,context)
+                                     (color       . ,color)))))
 
 (defun urgrep--rgrep-process-setup ()
   "Set up environment variables for rgrep.
 See also `grep-process-setup'."
-  ;; `setenv' modifies `process-environment' let-bound in `compilation-start'
+  ;; `setenv' modifies `process-environment' let-bound in `compilation-start'.
   ;; Any TERM except "dumb" allows GNU grep to use `--color=auto'.
   (setenv "TERM" "emacs-urgrep")
   ;; GREP_COLOR is used in GNU grep 2.5.1, but deprecated in later versions.
@@ -317,18 +355,19 @@ See also `grep-process-setup'."
      (executable-name . "ugrep")
      (regexp-syntax bre ere pcre)
      (arguments executable (:abbreviate color "-rn" "--ignore-files")
-                hidden-file file-wildcards group context case-fold regexp "-e"
-                query directory)
+                hidden-file file-wildcard group context case-fold regexp "-e"
+                query root)
      (regexp-arguments ('bre  '("-G"))
                        ('ere  '("-E"))
                        ('pcre '("-P"))
                        (_     '("-F")))
      (case-fold-arguments ((pred identity) '("-i")))
      (hidden-file-arguments ((pred identity) '("--hidden")))
-     (file-wildcards-arguments
+     (file-wildcard-arguments
       ((and x (pred identity))
        (mapcar (lambda (i) (concat "--include=" i)) x)))
-     (directory-arguments (x x))
+     (root-arguments ('(".") nil)
+                     (x x))
      (group-arguments ((pred identity) '("--heading" "--break")))
      (context-arguments . ,urgrep--context-arguments)
      (color-arguments
@@ -338,15 +377,16 @@ See also `grep-process-setup'."
     (ripgrep
      (executable-name . "rg")
      (regexp-syntax pcre)
-     (arguments executable (:abbreviate color) hidden-file file-wildcards group
-                context case-fold regexp "--" query directory)
+     (arguments executable (:abbreviate color) hidden-file file-wildcard group
+                context case-fold regexp "--" query root)
      (regexp-arguments ('nil '("-F")))
      (case-fold-arguments ((pred identity) '("-i")))
      (hidden-file-arguments ((pred identity) '("--hidden")))
-     (file-wildcards-arguments
+     (file-wildcard-arguments
       ((and x (pred identity))
        (flatten-list (mapcar (lambda (i) (cons "-g" i)) x))))
-     (directory-arguments (x x))
+     (root-arguments ('(".") nil)
+                     (x x))
      (group-arguments ('nil '("--no-heading"))
                       (_    '("--heading")))
      (context-arguments . ,urgrep--context-arguments)
@@ -359,16 +399,17 @@ See also `grep-process-setup'."
     (ag
      (executable-name . "ag")
      (regexp-syntax pcre)
-     (arguments executable (:abbreviate color) hidden-file file-wildcards group
-                context case-fold regexp "--" query directory)
+     (arguments executable (:abbreviate color) hidden-file file-wildcard group
+                context case-fold regexp "--" query root)
      (regexp-arguments ('nil '("-Q")))
      (case-fold-arguments ('nil '("-s"))
                           (_    '("-i")))
      (hidden-file-arguments ((pred identity) '("--hidden")))
-     (file-wildcards-arguments
+     (file-wildcard-arguments
       ((and x (pred identity))
        (list "-G" (urgrep--wildcards-to-regexp x 'pcre))))
-     (directory-arguments (x x))
+     (root-arguments ('(".") nil)
+                     (x x))
      (group-arguments ('nil '("--nogroup"))
                       (_    '("--group")))
      (context-arguments . ,urgrep--context-arguments)
@@ -378,16 +419,17 @@ See also `grep-process-setup'."
     (ack
      (executable-name . "ack")
      (regexp-syntax pcre)
-     (arguments executable (:abbreviate color) hidden-file file-wildcards group
-                context case-fold regexp "--" query directory)
+     (arguments executable (:abbreviate color) hidden-file file-wildcard group
+                context case-fold regexp "--" query root)
      (regexp-arguments ('nil '("-Q")))
      (case-fold-arguments ((pred identity) '("-i")))
      (hidden-file-arguments ('nil '("--ignore-dir=match:/^\\./"
                                     "--ignore-file=match:/^\\./")))
-     (file-wildcards-arguments
+     (file-wildcard-arguments
       ((and x (pred identity))
        (list "-G" (urgrep--wildcards-to-regexp x 'pcre))))
-     (directory-arguments (x x))
+     (root-arguments ('(".") nil)
+                     (x x))
      (group-arguments ('nil '("--nogroup"))
                       (_    '("--group")))
      (context-arguments . ,urgrep--context-arguments)
@@ -431,7 +473,7 @@ See also `grep-process-setup'."
                               "-c" "color.grep.separator="
                               "grep" "--color"))))
     (grep
-     ;; Note: We only use these for detecting the usability of find/grep. To
+     ;; Note: We only use these for detecting the usability of find/grep.  To
      ;; modify the programs that actually run, change `grep-find-template'.
      (executable-name "find" "grep")
      (regexp-syntax bre ere pcre)
@@ -448,10 +490,22 @@ See also `grep-process-setup'."
      (context-arguments . ,urgrep--context-arguments)))
   "An alist of known tools to try when running urgrep.")
 
+(defvar urgrep--cached-tool nil
+  "The cached urgrep tool to use.
+This value is connection-local.")
+
+(connection-local-set-profile-variables
+ 'urgrep-connection-local-profile
+ '((urgrep--cached-tool . nil)))
+
+(connection-local-set-profiles
+ '(:application tramp)
+ 'urgrep-connection-local-profile)
+
 (defcustom urgrep-preferred-tools nil
   "List of urgrep tools to search for.
 This can be nil to use the default list of tools in `urgrep-tools'
-or a list of tools to try in descending order of preference. Each
+or a list of tools to try in descending order of preference.  Each
 tool can be either a symbol naming the tool or a cons cell of the
 tool name and the file name of the executable (or a list thereof
 if there are multiple exeuctables)."
@@ -464,19 +518,10 @@ if there are multiple exeuctables)."
                      (append tool-choice
                              `((cons :tag "(tool . path)"
                                      ,tool-choice (string :tag "Path")))))))
-    :group 'urgrep)
-
-(defvar urgrep--cached-tool nil
-  "The cached urgrep tool to use.
-This value is connection-local.")
-
-(connection-local-set-profile-variables
- 'urgrep-connection-local-profile
- '((urgrep--cached-tool . nil)))
-
-(connection-local-set-profiles
- '(:application tramp)
- 'urgrep-connection-local-profile)
+  :set (lambda (symbol value)
+         (setq urgrep--cached-tool nil)
+         (set-default symbol value))
+  :group 'urgrep)
 
 (defsubst urgrep-connection-local-profile ()
   "Get a connection-local profile name for urgrep."
@@ -492,7 +537,8 @@ If SUFFIX is non-nil, append it to PROP to generate the property name."
 
 (defun urgrep--get-prop-pcase (prop tool value &optional suffix)
   "Get the property PROP from TOOL and use it as a `pcase' macro for VALUE.
-If SUFFIX is non-nil, append it to PROP to generate the property name."
+If SUFFIX is non-nil, append it to PROP to generate the property
+name."
   (when-let ((cases (urgrep--get-prop prop tool suffix))
              (block (append `(,#'pcase ',value) cases)))
     (eval block t)))
@@ -518,45 +564,44 @@ each tool, possibly modified with the executable path defined in
   "Get the preferred urgrep tool from `urgrep-tools'.
 This caches the default tool per-host in `urgrep--host-defaults'."
   (with-connection-local-variables
-   (or urgrep--cached-tool
-       (let ((vc-backend-name)
-             (saw-vc-tool-p nil))
-         (cl-loop
-          for tool iter-by (urgrep--iter-tools) do
-          (let ((tool-executable (urgrep--get-prop 'executable-name tool))
-                (tool-vc-backend (urgrep--get-prop 'vc-backend tool)))
-            (setq saw-vc-tool-p (or saw-vc-tool-p tool-vc-backend))
-            ;; Cache the VC backend name if we need it.
-            (when-let (((and tool-vc-backend (not vc-backend-name)))
-                       (proj (project-current)))
-              (setq vc-backend-name (vc-responsible-backend
-                                     (project-root proj))))
-            ;; If we find the executable (and it's for the right VC
-            ;; backend, if relevant), cache it and then return it.
-            (when (and (seq-every-p (lambda (i) (executable-find i t))
-                                    (ensure-list tool-executable))
-                       (or (not tool-vc-backend)
-                           (string= vc-backend-name tool-vc-backend)))
-              ;; So long as we didn't examine a VC-specific tool, we can
-              ;; cache this result for future calls, since the result will
-              ;; always be the same.  If we *did* see a VC-specific tool,
-              ;; this host will use different tools for different
-              ;; directories, so we can't cache anything.
-              (unless saw-vc-tool-p
-                (setq urgrep--cached-tool tool)
-                (when (file-remote-p default-directory)
-                  (connection-local-set-profile-variables
-                   (urgrep-connection-local-profile)
-                   `((urgrep--cached-tool . ,urgrep--cached-tool)))
-                  (connection-local-set-profiles
-                   (connection-local-criteria-for-default-directory)
-                   (urgrep-connection-local-profile))))
-              (cl-return tool))))))))
+   (let ((use-cache (equal urgrep-preferred-tools
+                           (default-value 'urgrep-preferred-tools)))
+         vc-backend-name)
+     (if (and urgrep--cached-tool use-cache)
+         urgrep--cached-tool
+       (catch 'found
+         (iter-do (tool (urgrep--iter-tools))
+           (let ((tool-executable (urgrep--get-prop 'executable-name tool))
+                 (tool-vc-backend (urgrep--get-prop 'vc-backend tool)))
+             ;; If we see a VC-specific tool, this host might use different
+             ;; tools for different directories, so we can't cache anything.
+             (setq use-cache (and use-cache (not tool-vc-backend)))
+             ;; Cache the VC backend name if we need it.
+             (when-let (((and tool-vc-backend (not vc-backend-name)))
+                        (proj (project-current)))
+               (setq vc-backend-name (vc-responsible-backend
+                                      (project-root proj))))
+             ;; If we find the executable (and it's for the right VC backend, if
+             ;; relevant), cache it if possible and then return it.
+             (when (and (seq-every-p (lambda (i) (executable-find i t))
+                                     (ensure-list tool-executable))
+                        (or (not tool-vc-backend)
+                            (string= vc-backend-name tool-vc-backend)))
+               (when use-cache
+                 (setq urgrep--cached-tool tool)
+                 (when (file-remote-p default-directory)
+                   (connection-local-set-profile-variables
+                    (urgrep-connection-local-profile)
+                    `((urgrep--cached-tool . ,urgrep--cached-tool)))
+                   (connection-local-set-profiles
+                    (connection-local-criteria-for-default-directory)
+                    (urgrep-connection-local-profile))))
+               (throw 'found tool)))))))))
 
 (defun urgrep-get-tool (&optional tool)
   "Get the urgrep tool for TOOL.
-If TOOL is nil, get the default tool.  If TOOL is a symbol, look it up
-in `urgrep-tools'.  Otherwise, return TOOL as-is."
+If TOOL is nil, get the default tool.  If TOOL is a symbol, look
+it up in `urgrep-tools'.  Otherwise, return TOOL as-is."
   (pcase tool
     ('nil (urgrep--get-default-tool))
     ((and (pred symbolp) tool) (assq tool urgrep-tools))
@@ -589,38 +634,47 @@ in `urgrep-tools'.  Otherwise, return TOOL as-is."
 
 ;;;###autoload
 (cl-defun urgrep-command (query &key tool regexp (case-fold 'inherit) hidden
-                                files directory (group t) (context 0) (color t))
+                                file-wildcard (root ".") (group t) (context 0)
+                                (color t))
   "Return a command to use to search for QUERY.
 Several keyword arguments can be supplied to adjust the resulting
 command:
 
-TOOL: a tool from `urgrep-tools': a key-value pair from the list, just
-the key, or nil to use the default tool.
+TOOL: a tool from `urgrep-tools': a key-value pair from the list,
+just the key, or nil to use the default tool.
 
 REGEXP: the style of regexp to use for results; one of nil (fixed
 strings), `bre' (basic regexp), `ere' (extend regexp), `pcre'
-\(Perl-compatible regexp), or t (the default regexp style stored in
-`urgrep-regexp-syntax').
+\(Perl-compatible regexp), or t (the default regexp style stored
+in `urgrep-regexp-syntax').
 
-CASE-FOLD: determine whether QUERY is case-sensitive or not; possible
-values are as `urgrep-case-fold', defaulting to `inherit'.
+CASE-FOLD: determine whether QUERY is case-sensitive or not;
+possible values are as `urgrep-case-fold', defaulting to
+`inherit'.
 
 HIDDEN: non-nil to search in hidden files; defaults to nil.
 
-FILES: a wildcard (or list of wildcards) to limit the files searched.
+FILE-WILDCARD: a wildcard (or list of wildcards) to limit the
+files searched.
 
-GROUP: show results grouped by filename (t, the default), or if nil,
-prefix the filename on each result line.
+ROOT: a file/directory (or list thereof) to search in; by
+default, search within `default-directory'.  If nil, don't search
+anywhere; this is useful when making a template command, e.g. to
+use with \"xargs\".
 
-CONTEXT: the number of lines of context to show around results; either
-an integer (to show the same number of lines before and after) or a
-cons (to show CAR and CDR lines before and after, respectively).
+GROUP: show results grouped by filename (t, the default), or if
+nil, prefix the filename on each result line.
+
+CONTEXT: the number of lines of context to show around results;
+either an integer (to show the same number of lines before and
+after) or a cons (to show CAR and CDR lines before and after,
+respectively).
 
 COLOR: non-nil (the default) if the output should use color."
   (with-connection-local-variables
    (let* ((regexp-syntax (if (eq regexp t) urgrep-regexp-syntax regexp))
-          (files (if (listp files) files (list files)))
-          (directory (if (listp directory) directory (list directory)))
+          (file-wildcard (ensure-list file-wildcard))
+          (root (mapcar #'urgrep--safe-file-name (ensure-list root)))
           (tool (or (urgrep-get-tool tool)
                     (error "unknown tool %s" tool)))
           (tool-re-syntax (urgrep--get-best-syntax regexp-syntax tool))
@@ -634,26 +688,18 @@ COLOR: non-nil (the default) if the output should use color."
      ;; Build the command arguments.
      (if cmd-fun
          (funcall cmd-fun query :tool tool :regexp tool-re-syntax
-                  :case-fold case-fold :hidden hidden :files files
-                  :directory directory :group group :context context
-                  :color color)
-       (let ((arguments (urgrep--get-prop 'arguments tool))
-             (abbrev (urgrep--get-prop 'abbreviations tool))
-             (props `((executable . ,(urgrep--get-prop 'executable-name tool))
-                      (query . ,query)
-                      ,@(mapcar (pcase-lambda (`(,k . ,v))
-                                  (cons k (urgrep--get-prop-pcase
-                                           k tool v "-arguments")))
-                                `((regexp         . ,tool-re-syntax)
-                                  (case-fold      . ,case-fold)
-                                  (hidden-file    . ,hidden)
-                                  (file-wildcards . ,files)
-                                  (directory      . ,directory)
-                                  (group          . ,group)
-                                  (context        . ,context)
-                                  (color          . ,color))))))
-         (urgrep--flatten-arguments (cl-sublis props arguments)
-                                    abbrev))))))
+                  :case-fold case-fold :hidden hidden
+                  :file-wildcard file-wildcard :root root
+                  :group group :context context :color color)
+       (urgrep--interpolate-arguments query tool
+                                      `((regexp        . ,tool-re-syntax)
+                                        (case-fold     . ,case-fold)
+                                        (hidden-file   . ,hidden)
+                                        (file-wildcard . ,file-wildcard)
+                                        (root          . ,root)
+                                        (group         . ,group)
+                                        (context       . ,context)
+                                        (color         . ,color)))))))
 
 
 ;; urgrep-mode
@@ -686,8 +732,8 @@ If EDIT-COMMAND is non-nil, the search can be edited."
     (urgrep--start command query urgrep-current-tool)))
 
 (defvar-keymap urgrep-mode-map
-  ;; Don't inherit from `compilation-minor-mode-map',
-  ;; because that introduces a menu bar item we don't want.
+  ;; Don't inherit from `compilation-minor-mode-map', because that introduces a
+  ;; menu bar item we don't want.
   :parent special-mode-map
   "<mouse-2>"     #'compile-goto-error
   "<follow-link>" 'mouse-face
@@ -723,8 +769,8 @@ If EDIT-COMMAND is non-nil, the search can be edited."
      :help "Kill the currently running search process"]))
 
 (defvar urgrep-mode-tool-bar-map
-  ;; When bootstrapping, tool-bar-map is not properly initialized yet,
-  ;; so don't do anything.
+  ;; When bootstrapping, `tool-bar-map' is not properly initialized yet, so
+  ;; don't do anything.
   (when (keymapp tool-bar-map)
     (let ((map (copy-keymap tool-bar-map)))
       (define-key map [undo] nil)
@@ -762,10 +808,10 @@ If EDIT-COMMAND is non-nil, the search can be edited."
 (defun urgrep-mode--looking-at-context-line ()
   "Return t if looking at a grep-like context line.
 
-If so, this function sets the match data, with the first match group
-indicating the separator between the line number and the match, and
-the second group indicating the separator between the file name and
-line number."
+If so, this function sets the match data, with the first match
+group indicating the separator between the line number and the
+match, and the second group indicating the separator between the
+file name and line number."
   ;; Use the `urgrep-file-name' property set by `urgrep-filter' to reliably
   ;; detect if the result was printed in grouped or ungrouped format.
   (if (get-text-property (point) 'urgrep-file-name)
@@ -800,8 +846,8 @@ line number."
     ;; Highlight context lines of various flavors.
     ((lambda (limit)
        (unless (bolp) (forward-line))
-       ;; Search line-by-line until we're looking at something that
-       ;; looks like a context line.
+       ;; Search line-by-line until we're looking at something that looks like a
+       ;; context line.
        (catch 'found
          (while (< (point) limit)
            (when (urgrep-mode--looking-at-context-line)
@@ -816,9 +862,11 @@ line number."
 (defvar urgrep--column-end-adjustment
   (if (< emacs-major-version 28) 0 1)
   "Handle core Emacs changes to the column range for `compile-mode' matches.
-In Emacs 28+, the column range for matches is closed, but in previous
-versions, it's half-open.  Use this to adjust the value as needed in
-`urgrep--column-end'.  For more details on the change, see
+In Emacs 28+, the column range for matches is closed, but in
+previous versions, it's half-open.  Use this to adjust the value
+as needed in `urgrep--column-end'.
+
+For more details on the change, see
 <https://debbugs.gnu.org/cgi/bugreport.cgi?bug=49624>.")
 
 (defun urgrep--column-begin ()
@@ -883,9 +931,9 @@ See `compilation-error-regexp-alist' for format details.")
 (defun urgrep-exit-message (status code msg)
   "Return a status message for urgrep results."
   (if (eq status 'exit)
-      ;; This relies on the fact that `compilation-start'
-      ;; sets buffer-modified to nil before running the command,
-      ;; so the buffer is still unmodified if there is no output.
+      ;; This relies on the fact that `compilation-start' sets buffer-modified
+      ;; to nil before running the command, so the buffer is still unmodified if
+      ;; there is no output.
       (cond ((and (zerop code) (buffer-modified-p))
              (if (> urgrep-num-matches-found 0)
                  (cons (format (ngettext "finished with %d match found\n"
@@ -1151,7 +1199,7 @@ current search.  Outside the prompt, this sets the value for all
 future searches."
   (interactive
    (let ((enable-recursive-minibuffers t))
-                 (list (split-string (read-string "File wildcard: ")))))
+     (list (split-string (read-string "File wildcard: ")))))
   (setq urgrep-file-wildcards files)
   (when (window-minibuffer-p) (urgrep--update-search-prompt)))
 
@@ -1169,29 +1217,32 @@ future searches."
 (cl-defun urgrep--read-query (initial &key tool (regexp urgrep-search-regexp)
                                       (case-fold urgrep-case-fold)
                                       (hidden urgrep-search-hidden-files)
-                                      (files urgrep-file-wildcards)
+                                      (file-wildcard urgrep-file-wildcards)
                                       (group urgrep-group-matches)
                                       (context urgrep-context-lines)
                                       (default (and (not initial)
                                                     (urgrep--search-default))))
   "Prompt the user for a search query starting with an INITIAL value.
 Return a list that can be passed to `urgrep-command' to turn into
-a shell command. TOOL, REGEXP, CASE-FOLD, FILES, GROUP, and
+a shell command.  TOOL, REGEXP, CASE-FOLD, FILES, GROUP, and
 CONTEXT are as in `urgrep-command'.
 
 In addition, you can pass DEFAULT to specify the default search
-query. If nil, guess the default based on the current region or
+query.  If nil, guess the default based on the current region or
 point."
+  ;; Get the tool in the current buffer, in case `urgrep-preferred-tools' has a
+  ;; buffer-local value.
+  (setq tool (urgrep-get-tool tool))
   ;; Run this in a temporary buffer to make sure that none of the dynamic
-  ;; variables below that we let-bind have buffer-local bindings. If they did,
+  ;; variables below that we let-bind have buffer-local bindings.  If they did,
   ;; we wouldn't be able to retrieve the values for them that we set from inside
   ;; the minibuffer.
   (with-temp-buffer
-    (let* ((urgrep-current-tool (urgrep-get-tool tool))
+    (let* ((urgrep-current-tool tool)
            (urgrep-search-regexp regexp)
            (urgrep-case-fold case-fold)
            (urgrep-search-hidden-files hidden)
-           (urgrep-file-wildcards files)
+           (urgrep-file-wildcards file-wildcard)
            (urgrep-context-lines context)
            (prompt (urgrep--search-prompt default))
            (query (minibuffer-with-setup-hook
@@ -1201,7 +1252,7 @@ point."
            (query (if (equal query "") default query)))
       (list query :tool urgrep-current-tool :regexp urgrep-search-regexp
             :case-fold urgrep-case-fold :hidden urgrep-search-hidden-files
-            :files urgrep-file-wildcards :group group
+            :file-wildcard urgrep-file-wildcards :group group
             :context urgrep-context-lines))))
 
 (defun urgrep--read-command (command)
@@ -1265,7 +1316,7 @@ searched."
   "Recursively search using the given COMMAND.
 
 TOOL is the search tool corresponding to COMMAND (see
-`urgrep-get-tool'). If nil, guess the tool based on the value of
+`urgrep-get-tool').  If nil, guess the tool based on the value of
 COMMAND.
 
 When called interactively, this behaves like `urgrep', but allows you
@@ -1333,11 +1384,11 @@ This is meant to be used as a command in Eshell."
          "number of lines of trailing context to print")
      ;; General options
      (?h "help" nil nil "show this help message")
-     :usage "[OPTION]... PATTERN [PATH]...
-Recursively search for PATTERN within PATH.")
+     :usage "[OPTION]... PATTERN [ROOT]...
+Recursively search for PATTERN within ROOT.")
    (unless args (error "Expected a search pattern"))
    (let* ((query (car args))
-          (directory (cdr args))
+          (root (cdr args))
           (context
            (cond
             (context-around (string-to-number context-around))
@@ -1346,7 +1397,7 @@ Recursively search for PATTERN within PATH.")
                    (if context-after (string-to-number context-after) 0)))))
           options)
      ;; Fill the options to pass to `urgrep'.
-     (when directory (setq options `(:directory ,directory       . ,options)))
+     (when root      (setq options `(:root      ,root            . ,options)))
      (when context   (setq options `(:context   ,context         . ,options)))
      (when hidden    (setq options `(:hidden    ,(car hidden)    . ,options)))
      (when group     (setq options `(:group     ,(car group)     . ,options)))
